#include	"GlutRotateWindow.h"
#include	"Program.h"
#include	"VertexBuffer.h"
#include	"VertexArray.h"

#define	NUM_VERTICES	3
#define	VERTEX_SIZE		(3*sizeof(float))

static const float vertices [] = 
{
	 -2.0f, -2.0f, 0.0f, 
	  0.0f,  2.0f, 0.0f, 
	  2.0f,  0.0f, 0.0f
};

class	TestWindow : public GlutRotateWindow
{
	Program	program;
	VertexArray		vao;
	VertexBuffer	buf;

public:
	TestWindow () : GlutRotateWindow ( 100, 100, 900, 900, "Visibility buffer: forward rendering" ) 
	{
		if ( !program.loadProgram ( "forward.glsl" ) )
			exit ( "Error building program: %s\n", program.getLog ().c_str () );

		program.bind ();

		vao.create  ();
		vao.bind    ();
		buf.create  ();
		buf.bind    ( GL_ARRAY_BUFFER );
		buf.setData ( NUM_VERTICES * VERTEX_SIZE, vertices, GL_STATIC_DRAW );

		program.setAttrPtr ( "pos", 3, VERTEX_SIZE, nullptr );

		buf.unbind     ();
		vao.unbind     ();
		program.unbind ();
	}

	void	redisplay () override
	{
		glm::mat4	mv   = getRotation  ();
		glm::mat4	proj = getProjection ();
	
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		program.bind             ();
		program.setUniformMatrix ( "mv",    mv   );
		program.setUniformMatrix ( "proj",  proj );

		vao.bind ();
	
		glDrawArrays ( GL_TRIANGLES, 0, NUM_VERTICES );
	
		vao.unbind ();
		program.unbind ();
	}

	void mouseWheel ( int wheel, int dir, int x, int y ) override
	{
		eye += glm::vec3 ( dir, dir, dir ) * 0.5f;
		
					// since eye value has changed
		reshape ( getWidth (), getHeight () );
		
		glutPostRedisplay ();
	}
};

int main ( int argc, char * argv [] )
{
	GlutWindow::init ( argc, argv );

	TestWindow win;

	return GlutWindow::run ();
}


