#include	"GlutRotateWindow.h"
#include	"Program.h"
#include	"VertexBuffer.h"
#include	"VertexArray.h"
#include	"Framebuffer.h"
#include	"ScreenQuad.h"
#include	"Texture.h"

#define	NUM_VERTICES	3
#define	VERTEX_SIZE		(6*sizeof(float))

static const float vertices [] = 
{
	 -2.0f, -2.0f, 0.0f, 1.0f, 0.0f, 0.0f,
	 -2.0f,  2.0f, 0.0f, 1.0f, 0.0f, 1.0f,
	  2.0f, -2.0f, 0.0f, 1.0f, 1.0f, 0.0f
};

static const char * modeNames [] =
{
	"baryucentrics",
	"ddx delta",
	"ddy delta",
	"true rho",
	"computed rho"
};

std::string captionMain ( "VertexBuffer: test texture gradients - " );

class	VisibilityBufferWindow : public GlutRotateWindow
{
	Program	        program;
	Program         visCreate;
	Program         gbufCreate;
	VertexArray		vao;
	VertexBuffer	buf;
	FrameBuffer     fb;		// visibility buffer FB
	ScreenQuad	    quad;
	Texture		    visBuffer;
	Texture		    decalMap;
	int             mode     = 0;
	int             numModes = std::size ( modeNames );

public:
	VisibilityBufferWindow () : GlutRotateWindow ( 100, 100, 900, 900, captionMain.c_str () ) 
	{
		decalMap.load2D ( "../../Textures/wood.png" );

		if ( !visBuffer.create2D ( getWidth (), getHeight (), GL_RED_INTEGER, GL_R32UI ) )
			exit ( "Error creating visibility buffer" );

        visBuffer.bind       ();
        visBuffer.setFilter  ( GL_NEAREST, GL_NEAREST );
        visBuffer.setWrapAll ( GL_CLAMP_TO_EDGE       );
        visBuffer.unbind     ();
        
            // vis buffer FB - visibility buffer + debug RGBAF16 buffer
		fb.create             ( getWidth (), getHeight (), FrameBuffer::depth24 );
		fb.bind               ();
		fb.attachColorTexture ( &visBuffer );
		
		if ( !fb.isOk () )
			exit ( "Error with fb\n" );
			
		fb.unbind ();
		
		if ( !visCreate.loadProgram ( "vis-buffer.glsl" ) )
			exit ( "Error building program: %s\n", visCreate.getLog ().c_str () );

		if ( !gbufCreate.loadProgram ( "g-buffer.glsl" ) )
			exit ( "Error building program: %s\n", gbufCreate.getLog ().c_str () );

		if ( !program.loadProgram ( "forward.glsl" ) )
			exit ( "Error building program: %s\n", program.getLog ().c_str () );

		visCreate.bind ();

		vao.create  ();
		vao.bind    ();
		buf.create  ();
		buf.bind    ( GL_ARRAY_BUFFER );
		buf.setData ( NUM_VERTICES * VERTEX_SIZE, vertices, GL_STATIC_DRAW );

		visCreate.setAttrPtr ( 0, 4, VERTEX_SIZE, nullptr );
		visCreate.setAttrPtr ( 1, 2, VERTEX_SIZE, (void *) ( 4 * sizeof ( float ) ) );

		buf.unbind     ();
		vao.unbind     ();
        
		visCreate.unbind ();
        
		glClearColor ( 0, 0, 0, 1 );

        setCaption ( captionMain + modeNames [mode] );
	}

	void	redisplay () override
	{
		glm::mat4	mv   = getRotation   ();
		glm::mat4	proj = getProjection ();
	    
            // render to visibility buffer
		fb.bind        ();	
		fb.drawBuffers ( 1 );
        
		glClear        ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		visCreate.bind             ();
		visCreate.setUniformMatrix ( "mv",    mv   );
		visCreate.setUniformMatrix ( "proj",  proj );

		vao.bind ();
	
		glDrawArrays ( GL_TRIANGLES, 0, NUM_VERTICES );
	
		vao.unbind ();
		visCreate.unbind ();
		fb.unbind ();
        
            // now render using visibility buffer
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        
        gbufCreate.bind ();
        gbufCreate.setUniformMatrix ( "mv",      mv   );
        gbufCreate.setUniformMatrix ( "proj",    proj );
        gbufCreate.setTexture       ( "visMap",  0    );
        gbufCreate.setTexture       ( "decalMap", 1    );
        
        fb.getColorBuffer ( 0 ) -> bind ( 0 );      // bind visibility buffer
        decalMap.bind ( 1 );      // by test map with uv
        
        buf.bindBase ( GL_SHADER_STORAGE_BUFFER, 0 );

        quad.render ();
        
	}

	void	keyTyped ( unsigned char key, int modifiers, int x, int y ) override
	{
		GlutRotateWindow :: keyTyped ( key, modifiers, x, y );

		if ( key == '+' )
			mode++;
		else
		if ( key == '-' )
			mode--;
        
        if ( mode < 0 )
            mode += numModes;
        
        if ( mode >= numModes )
            mode -= numModes;
        
        setCaption ( std::string ( "VertexBuffer: " ) + modeNames [mode] );
	}

	void mouseWheel ( int wheel, int dir, int x, int y ) override
	{
		eye += glm::vec3 ( dir, dir, dir ) * 0.5f;
		
					// since eye value has changed
		reshape ( getWidth (), getHeight () );
		
		glutPostRedisplay ();
	}
};

int main ( int argc, char * argv [] )
{
	GlutWindow::init ( argc, argv );

	VisibilityBufferWindow win;

	return GlutWindow::run ();
}


